/*
 * $RCSfile: ihid.c,v $  $Revision: 1.6 $  $Name:  $
 * $Id: ihid.c,v 1.6 2008/05/09 04:30:12 bpaauwe Exp $
 * $Author: bpaauwe $
 * $Date: 2008/05/09 04:30:12 $
 * ----------------------------------------------------------------------------
 *
 *  Copyright (c) Bob Paauwe (2007)
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 *  more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * ----------------------------------------------------------------------------
 *
 *  Description:
 *
 *  Device dependent Implimentation code for communications with an USB
 *  Insteon PLC via the HIDDEV interface.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/hiddev.h>

#define DEBUG 0
#include "idrv.h"

int hid_cts_sleep = 120; /* 120 usecs to sleep between USB reads */

void *ihid_init(char *dev);
void ihid_close(iusb_t *iplc);
int ihid_read(iusb_t *iplc);
int ihid_write(iusb_t *iplc, unsigned char *pkt, int len);
int ihid_write_echo(iusb_t *iplc, unsigned char *pkt, int rt);
int ihid_cts(iusb_t *iplc, int cts);

static int ihid_read_usb (iusb_t *iplc, unsigned char *pkt, int *count);

/*
 * ihid_init
 *
 * This function initializes the USB connection to the PLC. This must
 * be called before any other function since the handle returned is
 * passed as an argument to all other functions.
 */
void *ihid_init(char *dev)
{
	int *fd;
	struct hiddev_devinfo info;

	fd = (int *)malloc(sizeof(int));

	*fd = open (dev, O_NONBLOCK | O_RDWR);

	if (*fd < 0) {
		fprintf(stderr, "Error opening hiddev device %s\n", dev);
		free(fd);
		return NULL;
	}

	/*
	 * Verify that this is a PLC by checking the vendor and product
	 * Vendor should be 0x1bf
	 * Product should be 4 (are the more product id's?
	 */
	ioctl(*fd, HIDIOCGDEVINFO, &info);
	if (info.vendor != 0x10bf) { /* Smarthome */
		fprintf(stderr, "Vendor is not Smarthome, so probably not a PLC\n");
		close (*fd);
		free(fd);
		return NULL;
	}

	return (void *)fd;
}


/*
 * ihid_close
 *
 * Release control of the PLC and shutdown the connection to it.
 */
void ihid_close(iusb_t *iplc)
{
	int *fd;

	/* Cast the interface descriptor into the proper type (file descriptor) */
	fd = (int *)iplc->vfd;

	/* Close the file descriptor */
	close(*fd);
	free(fd);
}


/*
 * ihid_read       -- Read packets from the USB interface and buffer.
 * ihid_write      -- Write a data packet to the USB interface
 * ihid_write_echo -- Write a data packet to the USB interface
 * ihid_cts        -- Wait for a clear-to-send from the USB interface.
 *
 */

/*
 * ihid_read
 *
 * This reads packets from the USB driver interface and returns 1 byte at
 * a time to the caller.
 *
 * Parameters:
 *   iusb_t interface handle
 *
 * Returns:
 *   success: byte read
 *   error:   -110
 *
 * This might seem overly complex, but reads from the hid device are always
 * 8 byte packets, but the packet may have between 0 and 7 bytes of data
 * in it.  By using a ring-buffer concept, this doesn't need to keep track
 * of all the state information to track variable length data in the 8
 * byte packet.
 */

#define URB_TIMEOUT 60 /* This is near the minimum, might have to increase it */

int ihid_read(iusb_t *iplc)
{
	unsigned char packet_buf[8];
	int cnt = 0;
	int i;
	int timeout = 0;
	unsigned char rb;
	int *fd;

	fd = (int *)iplc->vfd;
	/* Wrap ring buffer if necessary */
	if (iplc->ring_head == RING_SIZE) {
		iplc->ring_head = 0;
	}

	/* Check for empty ring buffer */
	if (iplc->ring_head == iplc->ring_tail) {  /* ring buffer empty */
		do {
			memset(packet_buf, 0, 8);
			ihid_read_usb(iplc, packet_buf, &cnt);
			for (i = 0; i < cnt; i++) {
				iplc->ring_buffer[iplc->ring_tail++] = packet_buf[i];
				if (iplc->ring_tail == RING_SIZE) {
					iplc->ring_tail = 0;
				}
			}
			timeout++;
			usleep(1000);
		} while ((cnt < 1) && (timeout < URB_TIMEOUT));
	}

	if (iplc->ring_head != iplc->ring_tail) {
		/* When poping a byte from the ring buffer, clear it */
		rb = iplc->ring_buffer[iplc->ring_head];
		iplc->ring_buffer[iplc->ring_head] = 0x00;
		iplc->ring_head++;
		return rb;
	} else {
		return -110;
	}
}

/*
 * ihid_cts
 *
 * Device dependent function that waits for the PCL to provide a CTS signal.
 *
 * Parameters:
 *   iusb_t interface handle
 *   once    If >0 then only loop once.
 *
 * We could end up locking up in this function.
 */
#define CTS_TIMEOUT 10000

int ihid_cts(iusb_t *iplc, int once)
{
	unsigned char usbp[8];
	int i;
	int cnt;
	int timeout = 0;
	int *fd;

	fd = (int *)iplc->vfd;

	do {
		memset(usbp, 0, 8);
		ihid_read_usb(iplc, usbp, &cnt);

		/* If count is > 0, then there is data in this packet */
		if (cnt) {
			for (i = 0; i < cnt; i++) {
				iplc->ring_buffer[iplc->ring_tail++] = usbp[i];
				if (iplc->ring_tail == RING_SIZE) {
					iplc->ring_tail = 0;
				}
			}
		}

		if (once) {
			return 0;
		} else {
			timeout++;
			usleep(hid_cts_sleep);
		}
	} while ((timeout < CTS_TIMEOUT) && (!iplc->cts_status));

	if (timeout >= CTS_TIMEOUT) {
		fprintf(stderr, "Timeout in ihid_cts (%d)\n", iplc->cts_status);
	}
	iplc->cts_status = 1;
	return 0;
}


/*
 * ihid_write_echo
 *
 * Device dependent function to send a two byte packet and wait for the
 * PLC echo it back. This should wait for the PLC to be clear-to-send.
 *
 * Parameters:
 *   iplc   interface handle.
 *   pkt    character buffer holding the characters to send.
 *   len    size of the expected response?
 *
 */
int ihid_write_echo(iusb_t *iplc, unsigned char *pkt, int len)
{
	int rs;
	int to = 100; /* number of loops before timing out */
	int *fd;
	int cnt;
	int i;
	unsigned char in_pkt[8];

	fd = (int *)iplc->vfd;

	/*
	 * Some test code to read for a bit to make sure the PLC doesn't
	 * have data it is trying to put in the ring.
	 */
	to = 100;
	while (--to) {
		usleep(1000);
		ihid_read_usb (iplc, in_pkt, &cnt);
		for (i = 0; i < cnt; i++) {
			iplc->ring_buffer[iplc->ring_tail++] = in_pkt[i];
			if (iplc->ring_tail == RING_SIZE) {
				iplc->ring_tail = 0;
			}
		}
	}


	/*
	 * Before writing to the PLC, check the CTS?  Is this maybe some of
	 * my problems?  Is there anyway to check for CTS independent of
	 * reading bytes?  Or do we need to keep track of the CTS status
	 * somehow?
	 */
	to = 10;
	while (!iplc->cts_status && to) {
		ihid_cts(iplc, 0);
		to--;
	}
	if (to == 0) {
		fprintf(stderr, "ihid_write_echo: Timout waiting for CTS\n");
	}

	ihid_write(iplc, pkt, len);
	iplc->cts_status = 0;
	usleep(2000); /* Give the PLC some time to respond */
	rs = 0;
	to = 20;
	do {
		memset(in_pkt, 0, 8);
		ihid_read_usb (iplc, in_pkt, &cnt);

		if (cnt > 1 ) {
			DBGMSG("recv'd: ");
			for (i = 0; i < cnt; i++) {
				DBGMSG("0x%02x ", in_pkt[i]);
			}
			DBGMSG("\n");
		}

		for (i = 0; i < cnt; i++) {
			if (pkt[rs+1] == in_pkt[i]) {
				rs++;
			}
			iplc->ring_buffer[iplc->ring_tail++] = in_pkt[i];
			if (iplc->ring_tail == RING_SIZE) {
				iplc->ring_tail = 0;
			}
		}
		if (rs == len) {
			return rs;
		}

		/*
		 * It seems like if it didn't get a response by now, it isn't
		 * going to get one so try resending?
		 */
		if (to == 10) {
			DBGMSG("write_echo: timeout on 0x%02x 0x%02x, rs = %d  0x%02x\n",
					pkt[1], pkt[2], rs, in_pkt[0]);
			ihid_write(iplc, pkt, len);
		}

		to--;
		usleep(3000);
	} while (to != 0);

	fprintf(stderr, "ihid_write_echo: Timeout waiting for PLC to echo data.\n");
	return -1;
}

/*
 * ihid_write
 *
 * Send bytes to the interface.  This function does not wait for the PLC
 * to echo back the data sent. But it should check to make sure the
 * PLC status says it is clear-to-send.
 *
 * Parameters:
 *   iplc     interface handle
 *   pkt      buffer with data to send
 *   len      number of bytes to send
 */
int ihid_write(iusb_t *iplc, unsigned char *pkt, int len)
{
	int *fd;
	int cnt;
	struct hiddev_usage_ref out;
	struct hiddev_report_info report;

	fd = (int *)iplc->vfd;

	if (!iplc->cts_status) {
		ihid_cts(iplc, 0);
	}

	/* Fill up the USB output buffer */
	for (cnt = 0; cnt < len + 1; cnt++) {
		out.report_type = HID_REPORT_TYPE_OUTPUT;
		out.report_id = HID_REPORT_ID_FIRST;
		out.field_index = 0;
		out.usage_index = cnt;
		out.usage_code = 0;
		out.value = pkt[cnt];
		ioctl(*fd, HIDIOCSUSAGE, &out);
	}

	/* Send the buffer to the device */
	report.report_type = HID_REPORT_TYPE_OUTPUT;
	report.report_id = HID_REPORT_ID_FIRST;
	report.num_fields = 0;
	ioctl(*fd, HIDIOCSREPORT, &report);

	iplc->cts_status = 0;

	return len;
}


/*
 * Read HID events from the USB device if any are available. This
 * returns immediatly if no events are ready.
 *
 * If a packet is ready, there should be 8 HID events, one for each
 * byte of the packet.
 *    event.hid will be set to the byte number of the packet.
 *    event.value will hold the byte value.
 *
 *    event.hid == 1  is the valid data count. It should be between
 *    0 and 7 (in the lower nibble).
 *
 * If no events are ready, does that mean that the PLC is ready
 * to receive data and cts should be 1?
 */
static int ihid_read_usb (iusb_t *iplc, unsigned char *pkt, int *count)
{
	struct hiddev_event event;
	int len = 0;
	int *fd;

	fd = (int *)iplc->vfd;

	*count = 0;
	while (read(*fd, &event, sizeof(struct hiddev_event)) > 0) {
		event.hid &= 0xff;

		if (event.hid == 1) { /* data is number of bytes in packet */
			if (event.value > 8) { /* Check for CTS */
				iplc->cts_status = 1;
			}
			*count = event.value & 0x0f;
		} else if (event.hid == 8) { /* last byte in packet */
			*pkt = event.value;
			len++;
			return iplc->cts_status;
		} else {
			*pkt = event.value;
			pkt++;
			len++;
		}

		if (len > 8) {
			printf("In ihid_read_usb: packet length exceeded!\n");
			return iplc->cts_status;
		}
	}

	return iplc->cts_status;
}
